import os
import pytest
from pathlib import Path

@pytest.fixture
def path_to_masks():
    path = str(Path(__file__).parents[0].parents[0]) + os.path.sep + 'deepmreye/masks/'
    return path

@pytest.fixture
def path_to_testdata():
    path = str(Path(__file__).parents[0]) + os.path.sep + 'data/'
    return path


""" Additional methods which did not earn its own space in the main methods. Maybe because they are more general and made for higher purposes. """
import numpy as np
import pandas as pd
from math import atan2
from scipy import ndimage
from sklearn.metrics import r2_score
import warnings
import tensorflow.keras.backend as K
from tensorflow.keras.callbacks import LearningRateScheduler
import tensorflow as tf


def augment_input(X, rotation=0, shift=0, zoom=0):
    """ Augments 3D images
    Inputs: 
    - X : Batch of 3D images
    - rotation : Rotation in degree
    - shift : Shift in pixels
    - zoom : Zoom in factor

    Outputs:
    - X : Augmented batch of 3D images
    """
    def scaleit(image, factor):
        with warnings.catch_warnings():  # Weird bug, should be fixed in next scipy version
            warnings.simplefilter("ignore")
            height, width, tiefe, depth = image.shape
            zheight = int(np.round(factor * height))
            zwidth = int(np.round(factor * width))
            ztiefe = int(np.round(factor * tiefe))
            zdepth = depth

            if factor < 1.0:
                newimg = np.zeros_like(image)
                row = (height - zheight) // 2
                col = (width - zwidth) // 2
                layer = (tiefe - ztiefe) // 2
                newimg[row:row+zheight, col:col+zwidth, layer:layer+ztiefe, :] = ndimage.zoom(image, (float(factor), float(
                    factor), float(factor), 1.0), order=0, mode='constant', cval=0, prefilter=False)[0:zheight, 0:zwidth, 0:ztiefe, :]

                return newimg

            elif factor > 1.0:
                row = (height - zheight) // 2
                col = (width - zwidth) // 2
                layer = (tiefe - ztiefe) // 2

                newimg = ndimage.zoom(image, (float(factor), float(factor), float(
                    factor), 1.0),  order=0, mode='constant', cval=0, prefilter=False)

                extrah = (newimg.shape[0] - height) // 2
                extraw = (newimg.shape[1] - width) // 2
                extrad = (newimg.shape[2] - tiefe) // 2
                newimg = newimg[extrah:extrah+height,
                                extraw:extraw+width, extrad:extrad+tiefe, :]

                return newimg

            else:
                return image

    def rotate_img(image1, max_angle, axis=[0, 1, 2]):
        if 0 in axis:
            # rotate along x-axis
            angle = np.random.uniform(-max_angle[0], max_angle[0])
            image1 = ndimage.rotate(image1, angle, mode='constant', cval=0, axes=(
                1, 2), reshape=False, prefilter=False, order=0)

        if 1 in axis:
            # rotate along y-axis
            angle = np.random.uniform(-max_angle[1], max_angle[1])
            image1 = ndimage.rotate(image1, angle, mode='constant', cval=0, axes=(
                0, 2), reshape=False, prefilter=False, order=0)

        if 2 in axis:
            # rotate along z-axis
            angle = np.random.uniform(-max_angle[2], max_angle[2])
            image1 = ndimage.rotate(image1, angle, mode='constant', cval=0, axes=(
                0, 1), reshape=False, prefilter=False, order=0)
        return image1
    # Rotate
    X = np.array([rotate_img(x, rotation, axis=[0, 1, 2]) for x in X])

    # Then shift
    X = np.array([ndimage.shift(x, shift=(np.random.randint(-shift, shift), np.random.randint(-shift, shift),
                                          np.random.randint(-shift, shift), 0), order=0, mode='constant', cval=0, prefilter=False) for x in X])
    # Then zoom
    X = np.array([scaleit(x, np.random.uniform(1-zoom, 1+zoom)) for x in X])

    return X


def mish(x):
    return x * K.tanh(K.softplus(x))


def step_decay_schedule(initial_lr=1e-4, decay_factor=0.9, num_epochs=50):
    def schedule(epoch):
        return initial_lr * (1 - epoch / num_epochs) ** decay_factor
    return LearningRateScheduler(schedule)


def euclidean_distance(y_true, y_pred):
    return np.sqrt(np.sum(np.square(y_true - y_pred), axis=-1))


def angle_between_points(y_true, y_pred):
    dx = y_pred[:, 0] - y_true[:, 0]
    dy = y_pred[:, 1] - y_true[:, 1]
    rads = np.array([atan2(dyy, dxx) for dyy, dxx in zip(dy, dx)])
    rads += np.pi
    return rads


def get_model_scores(real_y, pred_y, euc_pred, **args):
    try:
        (agg_scores, subtr_scores) = quantify_predictions(real_y, pred_y, euc_pred, percentile_cut=None)
        (agg_scores_pct, subtr_scores_pct) = quantify_predictions(real_y, pred_y, euc_pred, **args)
    except ValueError:
        # Participant has only NaNs or no data, return empty dataframes
        agg_scores, subtr_scores, agg_scores_pct, subtr_scores_pct = np.random.rand(9)*np.NaN, np.random.rand(9)*np.NaN, np.random.rand(9)*np.NaN, np.random.rand(9)*np.NaN # 9 return parameters
    df_scores = pd.DataFrame([agg_scores, subtr_scores, agg_scores_pct, subtr_scores_pct], index=['Default', 'Default subTR', 'Refined', 'Refined subTR'])
    df_scores.columns = pd.MultiIndex.from_tuples((('Pearson', 'X'), ('Pearson', 'Y'), ('Pearson', 'Mean'), ('R^2-Score', 'X'), ('R^2-Score', 'Y'), ('R^2-Score', 'Mean'),
                                                   ('Eucl. Error', 'Mean'), ('Eucl. Error', 'Median'), ('Eucl. Error', 'Std')))
    return df_scores


def quantify_predictions(y_true, y_pred, euc_pred, subtr_functor=np.median, percentile_cut=None):
    # Take care of NaN in XYs
    nan_indices = np.any(np.isnan(y_true), axis=(1, 2))
    y_true = y_true[~nan_indices, ...]
    y_pred = y_pred[~nan_indices, ...]
    euc_pred = euc_pred[~nan_indices, ...]

    if y_true.size < 1:
        raise ValueError('Participant has no ground truth data, nothing to quantify')

    # Aggregate across subTR values
    y_true_agg = subtr_functor(y_true, axis=1)
    y_pred_agg = subtr_functor(y_pred, axis=1)
    euc_pred_agg = subtr_functor(euc_pred, axis=1)

    # Use flattened array for subTR comparison
    y_true_flat = np.reshape(y_true, (y_true.shape[0] * y_true.shape[1], -1))
    y_pred_flat = np.reshape(y_pred, (y_pred.shape[0] * y_pred.shape[1], -1))
    euc_pred_flat = np.reshape(euc_pred, (euc_pred.shape[0] * euc_pred.shape[1], -1))

    agg_scores = calculate_scores(y_true_agg, y_pred_agg, euc_pred_agg, percentile_cut=percentile_cut)
    flat_scores = calculate_scores(y_true_flat, y_pred_flat, euc_pred_flat[..., 0], percentile_cut=percentile_cut)

    return (agg_scores, flat_scores)


def calculate_scores(y_true, y_pred, euc_pred, percentile_cut=None):
    if percentile_cut is not None:
        bad_indices = euc_pred > np.percentile(euc_pred, percentile_cut)
        y_true = y_true[~bad_indices, ...]
        y_pred = y_pred[~bad_indices, ...]
        euc_pred = euc_pred[~bad_indices, ...]

    pearson_x = np.corrcoef(y_true[:, 0], y_pred[:, 0])[0, 1]
    pearson_y = np.corrcoef(y_true[:, 1], y_pred[:, 1])[0, 1]
    pearson_mean = (pearson_x + pearson_y) / 2
    r2_x = r2_score(y_true[:, 0], y_pred[:, 0])
    r2_y = r2_score(y_true[:, 1], y_pred[:, 1])
    r2_mean = (r2_x + r2_y) / 2

    euc = euclidean_distance(y_true, y_pred)

    return (pearson_x, pearson_y, pearson_mean, r2_x, r2_y, r2_mean, np.mean(euc), np.median(euc), np.std(euc))


def smooth_signal(signal, N):
    """
    Simple smoothing by convolving a filter with 1/N.

    Parameters
    ----------
    signal : array_like
        Signal to be smoothed
    N : int
        smoothing_factor

    Returns
    -------
    signal : array_like
            Smoothed signal
    """
    # Preprocess edges
    signal = np.concatenate([signal[0:N], signal, signal[-N:]])
    # Convolve
    signal = np.convolve(signal, np.ones((N,))/N, mode='same')
    # Postprocess edges
    signal = signal[N:-N]

    return signal


class color:
    # From https://stackoverflow.com/questions/8924173/how-do-i-print-bold-text-in-python
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


# --------------------------------------------------------------------------------
# ------------------------Command Line Options------------------------------------
# --------------------------------------------------------------------------------
class Arg:
    def __init__(self, *args, **kwargs):
        self.cli = args
        self.kwargs = kwargs

CLI_OPTIONS = {
    "verbosity": Arg(
        '-v', '--verbose',
        help='Verbosity level',
        default=0,
    ),
    "gpu_id": Arg(
        '--gpu_id',
        help="Which GPU to use for training",
        metavar='gpu_id',
        default=''
    ), 
    "dataset_path": Arg(
        '--dataset_path',
        help="Path to dataset. Base folder is also dataset name.",
        metavar='dataset_path',
        default='./'
    ), 
    "weights_path": Arg(
        '--weights_path',
        help="Path to where weights should be stored.",
        metavar='weights_path',
        default='./weights/'
    ), 
    "datasets": Arg(
        '--datasets',
        help="If given only train subset of all datasets.",
        metavar='datasets',
        default=None)}



import numpy as np
import pandas as pd
from os.path import join

import tensorflow as tf
import tensorflow.keras.backend as K
from deepmreye import architecture
from deepmreye.util import util
from deepmreye.util import data_generator


def train_model(dataset, generators, opts, clear_graph=True, save=False, model_path='./', workers=4, use_multiprocessing=True, models=None, return_untrained=False, verbose=0):
    """Trains the model given a cross validation, hold out or leave one out generator, given model options

    Parameters
    ----------
    dataset : str
        Description of dataset, used for saving dataset to file
    generators : generator
        Cross validation, Hold out or Leave one out generator, yielding X,y pairs
    opts : dict
        Model options for training
    clear_graph : bool, optional
        If computational graph should be reset before each run, by default True
    save : bool, optional
        If model weights should be saved to file, by default False
    model_path : str, optional
        Filepath to where model weights should be stored, by default './'
    workers : int, optional
        Number of workers used when using multiprocessing, by default 4
    use_multiprocessing : bool, optional
        If multiprocessing should be used, can speed up training by 10x if data loader is bottleneck, by default True
    models : Keras Model instance, optional
        Can be provided if already trained model should be used instead of training a new one, by default None
    return_untrained : bool, optional
        If true, returns untrained but compiled model, by default False
    verbose : int, optional
        Verbosity level, by default 0

    Returns
    -------
    model : Keras Model
        Full model instance, used for training uncertainty estimate
    model_inference : Keras Model
        Model instance used for inference, provides uncertainty estimate (unsupervised model)
    """    
    # Clear session if needed
    if clear_graph:
        K.clear_session()
    if use_multiprocessing:
        tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
    else:
        workers = 1

    # Unpack generators
    (training_generator, testing_generator, single_testing_generators, single_testing_names,
     single_training_generators, single_training_names, full_testing_list, full_training_list) = generators

    # Test datagenerator and get representative X and y
    ((X, y), _) = next(training_generator)
    if verbose > 0:
        print('Input shape {}, Output shape {}'.format(X.shape, y.shape))
        print('Subjects in training set: {}, Subjects in test set: {}'.format(
            len(single_training_generators), len(single_testing_generators)))

    # Learning rate scheduler
    lr_sched = util.step_decay_schedule(initial_lr=opts['lr'], decay_factor=0.9, num_epochs=opts['epochs'])

    # Get model
    if models is None:
        model, model_inference = architecture.create_standard_model(X.shape[1::], opts)
    else:
        model, model_inference = models
    if return_untrained:
        tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.INFO)
        return (model, model_inference)

    # Train model
    if verbose > 1:
        print(model.summary(line_length=200))
    model.fit(training_generator, steps_per_epoch=opts['steps_per_epoch'], epochs=opts['epochs'], validation_data=testing_generator,
              validation_steps=opts['validation_steps'], callbacks=[lr_sched], use_multiprocessing=use_multiprocessing, workers=workers)

    # Save model weights
    if save:
        model_inference.save_weights(join(model_path, f'modelinference_{dataset}.h5'))
    if use_multiprocessing:
        tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.INFO)

    return (model, model_inference)


def evaluate_model(dataset, model, generators, save=False, model_path='./', model_description='', verbose=0, **args):
    """Evaluates model performance given model and generators used for training the model. Evaluates only on test set. 

    Parameters
    ----------
    dataset : str
        Description of dataset, used for saving dataset to file
    model : Keras Model
        Full model instance, used for training uncertainty estimate
    generators : generator
        Cross validation, Hold out or Leave one out generator, yielding X,y pairs
    save : bool, optional
        If true, save test set predictions to file, by default False
    model_path : str, optional
        Filepath to where model weights should be stored, by default './'
    model_description : str, optional
        Description of model used for saving the model evaluations, by default ''
    verbose : int, optional
        Verbosity level, by default 0

    Returns
    -------
    evaluation: dict
        Raw gaze coordinates, returned for each participant
    scores: pandas DataFrame
        Evaluation metrics for gaze coordinates (Pearson, R2-Score, Euclidean Error)
    """    
    (training_generator, testing_generator, single_testing_generators, single_testing_names,
     single_training_generators, single_training_names, full_testing_list, full_training_list) = generators
    evaluation, scores = dict(), dict()
    for idx, subj in enumerate(full_testing_list):
        X, real_y = data_generator.get_all_subject_data(subj)
        (pred_y, euc_pred) = model.predict(X, verbose=verbose-2, batch_size=16)
        evaluation[subj] = {'real_y': real_y, 'pred_y': pred_y, 'euc_pred': euc_pred}

        # Quantify predictions
        df_scores = util.get_model_scores(real_y, pred_y, euc_pred, **args)
        scores[subj] = df_scores

        # Print evaluation
        if verbose > 0:
            print(util.color.BOLD + '{} / {} - Model Performance for {}'.format(idx + 1, len(single_testing_names), subj) + util.color.END)
            if verbose > 1:
                pd.set_option('display.width', 120)
                pd.options.display.float_format = '{:.3f}'.format
                print(df_scores)
            else:
                print('Default: r={:.3f}, subTR: r={:.3f}, Euclidean Error: {:.3f}°'.format(
                    df_scores[('Pearson', 'Mean')]['Default'], df_scores[('Pearson', 'Mean')]['Default subTR'], df_scores[('Eucl. Error', 'Mean')]['Default']))
            print('\n')

    # Save dict
    if save:
        np.save(join(model_path, f'results{model_description}_{dataset}.npy'), evaluation)

    return (evaluation, scores)



from tensorflow.keras.layers import Conv3D, Activation, Add, UpSampling3D, Lambda, Dense, GaussianNoise, RepeatVector, AveragePooling3D, InputSpec, Input, Reshape, Flatten, Dropout, concatenate, Layer
from tensorflow.keras import initializers, regularizers, constraints
from tensorflow.keras.models import Model
import tensorflow.keras.backend as K
from tensorflow.keras import optimizers
import tensorflow as tf


def create_standard_model(input_shape, opts):
    """Creates convolutional model for training and inference

    Parameters
    ----------
    input_shape : list
        Input shape for each sample (X,Y,Z). Size given by shape of smallest eye mask
    opts : dict
        All model options used for creating and training the model. See util.model_opts for available options

    Returns
    -------
    model : Keras Model
        Full model instance, used for training uncertainty estimate
    model_inference : Keras Model
        Model instance used for inference, provides uncertainty estimate (unsupervised model)
    """
    # Create input layer and add optional noise
    input_layer = Input(input_shape)
    input_layer_noise = GaussianNoise(opts['gaussian_noise'])(input_layer)

    # Initial convolution + dropout
    x = conv3d_block(input_layer_noise, filters=opts['filters'],
                     kernel_size=opts['kernel'], strides=1, activation=opts['activation'])
    x = Dropout(opts['dropout_rate'])(x, training=opts['mc_dropout'])

    # Downsample to bottleneck layer, but keep skip layers
    x, skip_layers = downsample_block(x, filters=opts['filters'], depth=opts['depth'],
                                      multiplier=opts['multiplier'], groups=opts['groups'], activation=opts['activation'])

    # After this layer we split into segmentation and regression part
    bottleneck_layer = Flatten()(x)

    # Regression block
    out_regression = regression_block(bottleneck_layer, num_dense=opts['num_dense'], num_fc=opts['num_fc'],
                                      activation=opts['activation'], dropout_rate=opts['dropout_rate'], inner_timesteps=opts['inner_timesteps'], mc_dropout=opts['mc_dropout'])

    # Confidence for regression block
    out_confidence = confidence_block(
        bottleneck_layer, num_fc=opts['num_fc'], activation=opts['activation'], dropout_rate=opts['dropout_rate'], inner_timesteps=opts['inner_timesteps'], mc_dropout=opts['mc_dropout'])

    # Create model
    real_regression_shape = out_regression.shape.as_list()
    real_regression = Input(real_regression_shape[1::])
    model = Model(inputs=[input_layer, real_regression], outputs=[out_regression])
    model_inference = Model(inputs=input_layer, outputs=[out_regression, out_confidence])

    # Add losses
    loss_euclidean, loss_confidence = compute_standard_loss(out_confidence, real_regression, out_regression)
    model.add_loss(opts['loss_euclidean']*loss_euclidean)
    model.add_loss(opts['loss_confidence']*loss_confidence)

    # Compile it
    model.compile(optimizer=optimizers.Adam(opts['lr']))
    model.metrics.append(loss_euclidean)
    model.metrics_names.append("Euclidean loss")
    model.metrics.append(loss_confidence)
    model.metrics_names.append("Confidence loss")

    return model, model_inference

# --- adult blocks
def res_block(input_layer, filters, groups, activation):
    input_layer_res = conv3d_block(input_layer, filters=filters, kernel_size=1, strides=1, activation=activation)

    x = GroupNormalization(groups=groups, axis=-1)(input_layer)
    x = Activation(activation)(x)
    x = conv3d_block(x, filters=filters, kernel_size=3, strides=1, activation=activation)

    x = GroupNormalization(groups=groups, axis=-1)(x)
    x = Activation(activation)(x)
    x = conv3d_block(x, filters=filters, kernel_size=3, strides=1, activation=activation)

    out = Add()([x, input_layer_res])
    return out

def downsample_block(input_layer, filters, depth, multiplier, groups, activation):
    x = input_layer
    skip_layers = []
    for level_number in range(depth):
        n_level_filters = int(multiplier**level_number) * filters

        for level in range(0, level_number):
            x = res_block(x, filters=n_level_filters, groups=groups, activation=activation)
        # For segmentation save layer after res_blocks
        skip_layers.append(x)
        if level_number < (depth - 1):
            x = conv3d_block(x, filters=n_level_filters, kernel_size=3, strides=2, activation=activation)
    x = GroupNormalization(groups=groups, axis=-1)(x)
    x = Activation(activation)(x)

    return x, skip_layers

def regression_block(input_layer, num_dense, num_fc, activation, dropout_rate, inner_timesteps, mc_dropout, dense_out=2):
    x = RepeatVector(inner_timesteps)(input_layer)
    # Split timesteps so each gets its own weight
    all_xs = list()
    for i in range(0, inner_timesteps):
        x0 = Lambda(lambda x: x[:, i, :])(x)
        for d in range(0, num_dense):
            x0 = Dense(num_fc, activation=activation)(x0)
            x0 = Dropout(dropout_rate)(x0, training=mc_dropout)
        x0 = Dense(dense_out, activation='linear')(x0)
        x0 = Reshape((1, -1))(x0)
        all_xs.append(x0)

    out = concatenate(all_xs, axis=1)
    return out

def confidence_block(input_layer, num_fc, activation, dropout_rate, inner_timesteps, mc_dropout):
    out_conf = Dense(num_fc, activation=activation)(input_layer)
    out_conf = Dropout(dropout_rate)(out_conf, training=mc_dropout)
    out_conf = Dense(inner_timesteps, activation=activation)(out_conf)
    return out_conf

# --- baby blocks
def conv3d_block(input_layer, filters, kernel_size, strides, activation):
    if strides > 1:
        x = Conv3D(filters=filters, kernel_size=(kernel_size, kernel_size, kernel_size),
                   strides=(1, 1, 1), padding='same', activation=activation)(input_layer)
        x = AveragePooling3D()(x)
    else:
        x = Conv3D(filters=filters, kernel_size=(kernel_size, kernel_size, kernel_size),
                   strides=strides, padding='same', activation=activation)(input_layer)
    return x

def upsampling_block(input_layer, size=2):
    x = UpSampling3D(size=size)(input_layer)
    return x

# --- loss blocks
def compute_standard_loss(out_confidence, real_reg, pred_reg):
    loss_euclidean = euclidean_distance(real_reg, pred_reg)
    loss_confidence = mean_squared_error(loss_euclidean, out_confidence)

    return K.mean(loss_euclidean), K.mean(loss_confidence)

def euclidean_distance(y_true, y_pred):
    return tf.sqrt(K.sum(K.square(y_true - y_pred), axis=-1))

def mean_squared_error(y_true, y_pred):
    return K.square(y_true - y_pred)

# --- Custom Layers
# Group Norm --- from https://raw.githubusercontent.com/titu1994/Keras-Group-Normalization/master/group_norm.py
class GroupNormalization(Layer):
    """Group normalization layer

    Group Normalization divides the channels into groups and computes within each group
    the mean and variance for normalization. GN's computation is independent of batch sizes,
    and its accuracy is stable in a wide range of batch sizes

    # Arguments
        groups: Integer, the number of groups for Group Normalization.
        axis: Integer, the axis that should be normalized
            (typically the features axis).
            For instance, after a `Conv2D` layer with
            `data_format="channels_first"`,
            set `axis=1` in `BatchNormalization`.
        epsilon: Small float added to variance to avoid dividing by zero.
        center: If True, add offset of `beta` to normalized tensor.
            If False, `beta` is ignored.
        scale: If True, multiply by `gamma`.
            If False, `gamma` is not used.
            When the next layer is linear (also e.g. `nn.relu`),
            this can be disabled since the scaling
            will be done by the next layer.
        beta_initializer: Initializer for the beta weight.
        gamma_initializer: Initializer for the gamma weight.
        beta_regularizer: Optional regularizer for the beta weight.
        gamma_regularizer: Optional regularizer for the gamma weight.
        beta_constraint: Optional constraint for the beta weight.
        gamma_constraint: Optional constraint for the gamma weight.

    # Input shape
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    # Output shape
        Same shape as input.

    # References
        - [Group Normalization](https://arxiv.org/abs/1803.08494)
    """

    def __init__(self,
                 groups=32,
                 axis=-1,
                 epsilon=1e-5,
                 center=True,
                 scale=True,
                 beta_initializer='zeros',
                 gamma_initializer='ones',
                 beta_regularizer=None,
                 gamma_regularizer=None,
                 beta_constraint=None,
                 gamma_constraint=None,
                 **kwargs):
        super(GroupNormalization, self).__init__(**kwargs)
        self.supports_masking = True
        self.groups = groups
        self.axis = axis
        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_constraint = constraints.get(beta_constraint)
        self.gamma_constraint = constraints.get(gamma_constraint)

    def build(self, input_shape):
        dim = input_shape[self.axis]

        if dim is None:
            raise ValueError('Axis ' + str(self.axis) + ' of '
                             'input tensor should have a defined dimension '
                             'but the layer received an input with shape ' +
                             str(input_shape) + '.')

        if dim < self.groups:
            raise ValueError('Number of groups (' + str(self.groups) + ') cannot be '
                             'more than the number of channels (' +
                             str(dim) + ').')

        if dim % self.groups != 0:
            raise ValueError('Number of groups (' + str(self.groups) + ') must be a '
                             'multiple of the number of channels (' +
                             str(dim) + ').')

        self.input_spec = InputSpec(ndim=len(input_shape),
                                    axes={self.axis: dim})
        shape = (dim,)

        if self.scale:
            self.gamma = self.add_weight(shape=shape,
                                         name='gamma',
                                         initializer=self.gamma_initializer,
                                         regularizer=self.gamma_regularizer,
                                         constraint=self.gamma_constraint)
        else:
            self.gamma = None
        if self.center:
            self.beta = self.add_weight(shape=shape,
                                        name='beta',
                                        initializer=self.beta_initializer,
                                        regularizer=self.beta_regularizer,
                                        constraint=self.beta_constraint)
        else:
            self.beta = None
        self.built = True

    def call(self, inputs, **kwargs):
        input_shape = K.int_shape(inputs)
        tensor_input_shape = K.shape(inputs)

        # Prepare broadcasting shape.
        reduction_axes = list(range(len(input_shape)))
        del reduction_axes[self.axis]
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self.axis] = input_shape[self.axis] // self.groups
        broadcast_shape.insert(1, self.groups)

        reshape_group_shape = K.shape(inputs)
        group_axes = [reshape_group_shape[i] for i in range(len(input_shape))]
        group_axes[self.axis] = input_shape[self.axis] // self.groups
        group_axes.insert(1, self.groups)

        # reshape inputs to new group shape
        group_shape = [group_axes[0], self.groups] + group_axes[2:]
        group_shape = K.stack(group_shape)
        inputs = K.reshape(inputs, group_shape)

        group_reduction_axes = list(range(len(group_axes)))
        group_reduction_axes = group_reduction_axes[2:]

        mean = K.mean(inputs, axis=group_reduction_axes, keepdims=True)
        variance = K.var(inputs, axis=group_reduction_axes, keepdims=True)

        inputs = (inputs - mean) / (K.sqrt(variance + self.epsilon))

        # prepare broadcast shape
        inputs = K.reshape(inputs, group_shape)
        outputs = inputs

        # In this case we must explicitly broadcast all parameters.
        if self.scale:
            broadcast_gamma = K.reshape(self.gamma, broadcast_shape)
            outputs = outputs * broadcast_gamma

        if self.center:
            broadcast_beta = K.reshape(self.beta, broadcast_shape)
            outputs = outputs + broadcast_beta

        outputs = K.reshape(outputs, tensor_input_shape)

        return outputs

    def get_config(self):
        config = {
            'groups': self.groups,
            'axis': self.axis,
            'epsilon': self.epsilon,
            'center': self.center,
            'scale': self.scale,
            'beta_initializer': initializers.serialize(self.beta_initializer),
            'gamma_initializer': initializers.serialize(self.gamma_initializer),
            'beta_regularizer': regularizers.serialize(self.beta_regularizer),
            'gamma_regularizer': regularizers.serialize(self.gamma_regularizer),
            'beta_constraint': constraints.serialize(self.beta_constraint),
            'gamma_constraint': constraints.serialize(self.gamma_constraint)
        }
        base_config = super(GroupNormalization, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape



import os
import pytest
import numpy as np
from deepmreye import preprocess, train
from deepmreye.util import model_opts, data_generator, util

# --------------------------------------------------------------------------------
# --------------------------PREPROCESSING-----------------------------------------
# --------------------------------------------------------------------------------
def test_download(path_to_masks):
    # To only test this command python -m pytest -k 'download', all commands python -m pytest
    # Delete files if already in folder to see if download works
    for m in ['eyemask_small.nii', 'eyemask_big.nii', 'dme_template.nii']:
        if os.path.exists(path_to_masks + m):
            os.remove(path_to_masks + m)
    test_masks(path_to_masks)

def test_masks(path_to_masks):
    (eyemask_small, eyemask_big, dme_template, mask, x_edges, y_edges, z_edges) = preprocess.get_masks(path_to_masks)
    for m in [eyemask_small, eyemask_big, dme_template, mask]:
        np.testing.assert_equal(m.shape, (91, 109, 91))
    np.testing.assert_equal(x_edges, (41, 18, 73, 49))

def test_example_participant(path_to_masks, path_to_testdata):
    # Create eye mask for example participant
    path_to_participant = path_to_testdata + 'test_participant.nii'
    (eyemask_small, eyemask_big, dme_template, mask, x_edges, y_edges, z_edges) = preprocess.get_masks(path_to_masks)
    (masked_eye, transformation_statistics) = preprocess.run_participant(path_to_participant, dme_template, eyemask_big, eyemask_small, x_edges, y_edges, z_edges)
    np.testing.assert_equal(masked_eye.shape, (47, 29, 18, 2))
    
    # Ants transform is not fully deterministic so check for rough match of transformation
    assert(transformation_statistics[0] < -2)
    assert(transformation_statistics[1] < 0.5)
    assert(transformation_statistics[2] < 0.5)
    
    # Combine with labels
    this_mask = preprocess.normalize_img(masked_eye)
    this_label = preprocess.load_label(path_to_testdata, label_type='calibration_run')
    this_label = this_label[0: this_mask.shape[3], ...] # Adjust for testing
    this_id = ['test_participant']*this_label.shape[0], [0]*this_label.shape[0]

    preprocess.save_data('test_participant', [this_mask], [this_label], [this_id], path_to_testdata + 'processed/', center_labels=False)

def test_load_label(path_to_testdata):
    this_label = preprocess.load_label(path_to_testdata, label_type='calibration_run')
    np.testing.assert_equal(this_label.shape, (135, 10, 2))

# --------------------------------------------------------------------------------
# --------------------------MODEL TRAINING----------------------------------------
# --------------------------------------------------------------------------------
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_model_training(path_to_testdata):
    # Force test on CPU
    os.environ["CUDA_VISIBLE_DEVICES"] = "" 

    # Define test options. Use small values to speed up CI
    opts = model_opts.get_opts()
    opts['epochs'] = 2
    opts['steps_per_epoch'] = 1
    opts['validation_steps'] = 1

    # For test method use same participant in training and testset
    generators = data_generator.create_leaveoneout_generators([path_to_testdata + 'processed/', path_to_testdata + 'processed/'], batch_size=opts['batch_size'], augment_list=((opts['rotation_x'], opts['rotation_y'], opts['rotation_z']), opts['shift'], opts['zoom']), mixed_batches=True)
    
    # Train model
    (model, model_inference) = train.train_model(dataset='example_data', generators=generators[0], opts=opts, use_multiprocessing=True,
                                                return_untrained=False, verbose=1, save=False)

    # Evaluate model
    (evaluation, scores) = train.evaluate_model(dataset='example_data', model=model_inference, generators=generators[0], save=False, model_description='', verbose=2, percentile_cut=80)
    assert(isinstance(evaluation, dict))

def test_model_evaluation():
    # Generate artificial data
    num_points = 1000
    real_y = np.random.rand(num_points, 10, 2)
    pred_y = np.random.rand(num_points, 10, 2)
    euc_pred = np.random.rand(num_points, 10, 1)
    df_scores = util.get_model_scores(real_y, pred_y, euc_pred)
    
    # For all NaN participant
    real_y = np.random.rand(num_points, 10, 2) * np.NaN
    pred_y = np.random.rand(num_points, 10, 2) * np.NaN
    euc_pred = np.random.rand(num_points, 10, 1) * np.NaN
    df_scores = util.get_model_scores(real_y, pred_y, euc_pred)



import os
import numpy as np
import pandas as pd
import warnings

import plotly.graph_objs as go
from plotly.subplots import make_subplots

# --------------------------------------------------------------------------------
# --------------------------VISUALIZATIONS----------------------------------------
# --------------------------------------------------------------------------------

def visualise_input_data(X, y, color="rgb(0, 150, 175)", cut_at=151, bg_color="rgb(247,247,247)", ylim=[-6, 6], num_functionals=78):
    # Prepare data for plotting
    # For visualisation we use a downsampled and padded version of X and split the eye balls.
    X_right = np.pad(X[:, 0:25, ..., 0], ((0, 0), (0, 5), (1, 2), (0, 0)))
    X_left = np.pad(X[:, 25::, ..., 0], ((0, 0), (6, 2), (1, 2), (0, 0)))
    X_concat = np.concatenate((X_left, X_right), axis=1)
    X_vis = X_concat[0:cut_at, ...]

    # Output can contain NaN which is taken care of during model training. We just remove NaN here for visualization purposes
    vis_x = pd.Series(np.median(y[0:cut_at, ..., 0], axis=1)).ffill().values
    vis_y = pd.Series(np.median(y[0:cut_at, ..., 1], axis=1)).ffill().values

    # Create interactive plot to visualize example participant
    fig = make_subplots(rows=2, cols=4, horizontal_spacing=0.01, vertical_spacing=0.15,
                        specs=[[{"rowspan": 2, "colspan": 2}, None, {"colspan": 2}, None], [None, None, {"colspan": 2}, None]])
    fig.add_trace(go.Scatter(x=np.arange(0, len(vis_x)), y=vis_x, mode='lines', line_color='rgb(0,0,0)', opacity=0.5, line_width=3), row=1, col=3)
    fig.add_trace(go.Scatter(x=np.arange(0, len(vis_y)), y=vis_y, mode='lines', line_color='rgb(0,0,0)', opacity=0.5, line_width=3), row=2, col=3)

    # Plot input signal together with split output signal (X & Y)
    for i in range(0, X_vis.shape[0]):
        fig.add_trace(go.Scatter(x=[i], y=[vis_x[i]], mode='markers', marker=dict(size=12, color=color, line=dict(width=2, color='DarkSlateGrey')), visible=False), row=1, col=3)
        fig.add_trace(go.Scatter(x=[i], y=[vis_y[i]], mode='markers', marker=dict(size=12, color=color, line=dict(width=2, color='DarkSlateGrey')), visible=False), row=2, col=3)
        this_z = np.mean(X_vis[i, ..., :], axis=-1).transpose()
        if bg_color == "rgb(247,247,247)":
            colorscale = 'RdBu'
        elif bg_color == "rgb(255,255,255)":
            colorscale = 'RdGy'
        fig.add_heatmap(z=this_z, colorscale=colorscale, visible=False, showscale=False, zmid=0, row=1, col=1, name="TR: {}".format(i))  # colorbar=dict(x=0.45, y=0.5, thickness=10, len=0.5))
    fig.data[num_functionals*3+2].visible = True  # Some arithmetic;
    fig.data[num_functionals*3+2+1].visible = True  # connects to the active value below;
    fig.data[num_functionals*3+2+2].visible = True  # should be abstracted as a variable;

    # Add slider for changing TR
    steps, stepcount = [], 0
    for i in range(2, len(fig.data) - 2, 3):
        step = dict(method="update", label="TR: {}".format(stepcount), args=[{"visible": [True, True] + [False] * (len(fig.data) - 2)}])
        step["args"][0]["visible"][i] = True
        step["args"][0]["visible"][i+1] = True
        step["args"][0]["visible"][i+2] = True
        steps.append(step)
        stepcount += 1
    sliders = [dict(active=num_functionals, currentvalue={"prefix": "TR: ", "visible": False}, pad={"t": 40}, steps=steps)]

    # Add arrows
    annotations = [dict(x=11, y=8, xref="x", yref="y", text="Left eye", font=(dict(size=20)), showarrow=True, arrowhead=5, ax=-40, ay=80),
                   dict(x=46, y=8, xref="x", yref="y", text="Right eye", font=(dict(size=20)), showarrow=True, arrowhead=5, ax=40, ay=80),
                   dict(x=0.18, y=1.08, xref='paper', yref='paper', text="<b>Normalized MR-Signal</b>", font=(dict(size=20)), showarrow=False),
                   dict(x=0.78, y=1.08, xref='paper', yref='paper', text="<b>Gaze position</b>", font=(dict(size=20)), showarrow=False)]

    # Update layout and axes descriptions
    fig.update_layout(sliders=sliders, showlegend=False, margin=dict(t=70, l=20, b=20, r=20), autosize=False, width=1600, height=600,
                      plot_bgcolor=bg_color, paper_bgcolor=bg_color, annotations=annotations)
    fig.update_xaxes(showticklabels=False, row=1, col=1)  # title=dict(text='Eyeball signal', font=dict(size=50), standoff=0))
    fig.update_yaxes(showticklabels=False, row=1, col=1)
    fig.update_yaxes(range=ylim, row=1, col=3, ticksuffix='°', title=dict(text='X', standoff=0, font=dict(size=20)))
    fig.update_yaxes(range=ylim, row=2, col=3, ticksuffix='°', title=dict(text='Y', standoff=0, font=dict(size=20)))
    fig.update_xaxes(range=[-2, 150+2], row=1, col=3, title=dict(text='Functional Volume (TR)', standoff=16, font=dict(size=20)))
    fig.update_xaxes(range=[-2, 150+2], row=2, col=3)
    fig.update_layout()

    return fig


def visualise_predictions_click(evaluation, scores, color="rgb(0, 150, 175)", bg_color="rgb(247,247,247)"):
    # Prepare data for plotting
    all_scores = []
    for _, item in scores.items():
        all_scores.append(item.values)
    all_scores = np.array(all_scores)
    to_plot = np.concatenate((all_scores[..., 2], all_scores[..., 5]), axis=0)  # Pearson Mean & R2 Mean
    x = ['Pearson'] * all_scores[..., 2].shape[0] + ['R^2-Score'] * all_scores[..., 5].shape[0]
    participants = list(evaluation.keys())
    participants = participants * 2  # Repeat once for each statistic

    fig = go.FigureWidget(make_subplots(rows=2, cols=4, horizontal_spacing=0.01, vertical_spacing=0.15, shared_xaxes='columns',
                                        specs=[[{"rowspan": 2, "colspan": 2}, None, {"colspan": 2}, None], [None, None, {"colspan": 2}, None]]))

    fig.add_trace(go.Box(y=to_plot[:, 0], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Default', line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)
    fig.add_trace(go.Box(y=to_plot[:, 1], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Default subTR', line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)
    fig.add_trace(go.Box(y=to_plot[:, 2], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Refined', line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)
    fig.add_trace(go.Box(y=to_plot[:, 3], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Refined subTR', line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)

    this_real = np.median(evaluation[participants[0]]['real_y'], axis=1)
    this_pred = np.median(evaluation[participants[0]]['pred_y'], axis=1)
    fig.add_trace(go.Scatter(x=np.arange(0, len(this_real[:, 0])), y=this_real[:, 0], mode='lines', line_color='rgb(0,0,0)', opacity=0.5, line_width=3), row=1, col=3)
    fig.add_trace(go.Scatter(x=np.arange(0, len(this_pred[:, 0])), y=this_pred[:, 0], mode='lines', line_color=color, opacity=0.85, line_width=3), row=1, col=3)
    fig.add_trace(go.Scatter(x=np.arange(0, len(this_real[:, 1])), y=this_real[:, 1], mode='lines', line_color='rgb(0,0,0)', opacity=0.5, line_width=3), row=2, col=3)
    fig.add_trace(go.Scatter(x=np.arange(0, len(this_pred[:, 1])), y=this_pred[:, 1], mode='lines', line_color=color, opacity=0.85, line_width=3), row=2, col=3)

    # create our callback function
    def update_point(trace, points, selector):
        if points.point_inds:
            this_participant = participants[points.point_inds[0]]
            this_real = np.median(evaluation[this_participant]['real_y'], axis=1)
            this_pred = np.median(evaluation[this_participant]['pred_y'], axis=1)

            with fig.batch_update():
                all_scatterplots[0].x = np.arange(0, len(this_real[:, 0]))
                all_scatterplots[0].y = this_real[:, 0]
                all_scatterplots[1].x = np.arange(0, len(this_pred[:, 0]))
                all_scatterplots[1].y = this_pred[:, 0]

                all_scatterplots[2].x = np.arange(0, len(this_real[:, 1]))
                all_scatterplots[2].y = this_real[:, 1]
                all_scatterplots[3].x = np.arange(0, len(this_pred[:, 1]))
                all_scatterplots[3].y = this_pred[:, 1]
            fig.update_xaxes(range=[-2, 150+2], row=1, col=3, title=dict(text='Input Volume (TR)', standoff=16, font=dict(size=20)))
            fig.update_xaxes(range=[-2, 150+2], row=2, col=3)
    all_boxplots = fig.data[0:4]
    all_scatterplots = fig.data[4::]
    for bp in all_boxplots:
        bp.on_click(update_point)

    annotations = [dict(x=0.14, y=1.08, xref='paper', yref='paper', text="<b>Model Performance across participants</b>", font=(dict(size=20)), showarrow=False),
                   dict(x=0.855, y=1.08, xref='paper', yref='paper', text="<b>Predicted vs. True gaze position</b>", font=(dict(size=20)), showarrow=False)]

    fig.update_layout(showlegend=False, margin=dict(t=70, l=50, b=50, r=50), plot_bgcolor=bg_color, paper_bgcolor=bg_color,
                      boxmode='group', autosize=False, width=1600, height=650, annotations=annotations)
    fig.update_yaxes(range=[-1, 1], row=1, col=1)
    fig.update_xaxes(tickfont=dict(size=20), row=1, col=1)
    fig.update_yaxes(range=[-6, 6], row=1, col=3, ticksuffix='°', title=dict(text='X', standoff=0, font=dict(size=20)))
    fig.update_yaxes(range=[-6, 6], row=2, col=3, ticksuffix='°', title=dict(text='Y', standoff=0, font=dict(size=20)))
    fig.update_xaxes(range=[-2, 150+2], row=1, col=3, title=dict(text='Functional Volume (TR)', standoff=16, font=dict(size=20)))
    fig.update_xaxes(range=[-2, 150+2], row=2, col=3)

    return fig


def visualise_predictions_slider(evaluation, scores, color="rgb(0, 150, 175)", bg_color="rgb(247,247,247)", line_color="rgb(240,240,240)", ylim=[-6, 6], subTR=False):
    # Prepare data for plotting
    all_scores = []
    for _, item in scores.items():
        all_scores.append(item.values)
    all_scores = np.array(all_scores)
    to_plot = np.concatenate((all_scores[..., 2], all_scores[..., 5]), axis=0)  # Pearson Mean & R2 Mean
    x = ['Pearson'] * all_scores[..., 2].shape[0] + ['R^2-Score'] * all_scores[..., 5].shape[0]
    participants = list(evaluation.keys())
    hover_texts = []
    for subj in participants * 2:
        this_sub = os.path.splitext(os.path.basename(subj))[0]
        hover_texts.append('participant {}'.format(this_sub))
    # participants = participants * 4

    fig = go.FigureWidget(make_subplots(rows=2, cols=4, horizontal_spacing=0.05, vertical_spacing=0.15, shared_xaxes='columns',
                                        specs=[[{"rowspan": 2, "colspan": 2}, None, {"colspan": 2}, None], [None, None, {"colspan": 2}, None]]))

    fig.add_trace(go.Box(y=to_plot[:, 0], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Default', text=hover_texts, line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)
    fig.add_trace(go.Box(y=to_plot[:, 1], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Default subTR', text=hover_texts, line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)
    fig.add_trace(go.Box(y=to_plot[:, 2], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Refined', text=hover_texts, line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)
    fig.add_trace(go.Box(y=to_plot[:, 3], marker_size=12, x=x, boxpoints='all', pointpos=0, marker=dict(opacity=0.65, color=color, line=dict(color='rgb(0,0,0)', width=2)),
                         name='Refined subTR', text=hover_texts, line=dict(color='rgb(0,0,0)'), fillcolor="rgb(180, 180, 180)"
                         ), row=1, col=1)

    # Plot input signal together with split output signal (X & Y)
    for key, item in evaluation.items():
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if subTR:
                this_real = np.reshape(item['real_y'], (item['real_y'].shape[0] * item['real_y'].shape[1], -1))
                this_pred = np.reshape(item['pred_y'], (item['pred_y'].shape[0] * item['pred_y'].shape[1], -1))          
            else:
                this_real = np.nanmedian(item['real_y'], axis=1)
                this_pred = np.nanmedian(item['pred_y'], axis=1)
        this_sub = os.path.splitext(os.path.basename(key))[0]

        fig.add_trace(go.Scatter(x=np.arange(0, len(this_real[:, 0])), y=this_real[:, 0], mode='lines', visible=False,
                                 line_color='rgb(0,0,0)', opacity=0.5, line_width=3, name=this_sub), row=1, col=3)
        fig.add_trace(go.Scatter(x=np.arange(0, len(this_pred[:, 0])), y=this_pred[:, 0], mode='lines', visible=False,
                                 line_color=color, opacity=0.85, line_width=3, name=this_sub), row=1, col=3)
        fig.add_trace(go.Scatter(x=np.arange(0, len(this_real[:, 1])), y=this_real[:, 1], mode='lines', visible=False,
                                 line_color='rgb(0,0,0)', opacity=0.5, line_width=3, name=this_sub), row=2, col=3)
        fig.add_trace(go.Scatter(x=np.arange(0, len(this_pred[:, 1])), y=this_pred[:, 1], mode='lines', visible=False,
                                 line_color=color, opacity=0.85, line_width=3, name=this_sub), row=2, col=3)
    for i in range(0, 4):
        fig.data[4+i].visible = True

    # Add slider for changing participant
    steps, stepcount = [], 0
    for i in range(4, len(fig.data), 4):
        # print('Subs {}, i = {}, sub i-4: {}'.format(participants, i, participants[i-4]))
        this_sub = os.path.splitext(os.path.basename(participants[stepcount]))[0]
        step = dict(method="update", label="{}".format(this_sub), args=[{"visible": 4 * [True] + [False] * (len(fig.data) - 4)}])
        for j in range(0, 4):
            step["args"][0]["visible"][i+j] = True
        steps.append(step)
        stepcount += 1
    sliders = [dict(active=0, currentvalue={"prefix": "", "visible": False}, pad={"t": 70, "b": 10}, steps=steps)]  # len=0.95, x=0.05)]

    annotations = [dict(x=0.10, y=1.08, xref='paper', yref='paper', text="<b>Model Performance across participants</b>", font=(dict(size=20)), showarrow=False),
                   dict(x=0.855, y=1.08, xref='paper', yref='paper', text="<b>Predicted vs. True gaze position</b>", font=(dict(size=20)), showarrow=False)]

    fig.update_layout(showlegend=False, margin=dict(t=70, l=50, b=50, r=50), plot_bgcolor=bg_color, paper_bgcolor=bg_color,
                      boxmode='group', autosize=False, width=1600, height=650, annotations=annotations, sliders=sliders)

    fig.update_yaxes(range=[-1.1, 1.1], linecolor=line_color, zerolinecolor=line_color, gridcolor=line_color, row=1, col=1)
    fig.update_xaxes(tickfont=dict(size=20), row=1, col=1)
    fig.update_yaxes(range=ylim, row=1, col=3, linecolor=line_color, zerolinecolor=line_color, gridcolor=line_color, ticksuffix='°', title=dict(text='X', standoff=0, font=dict(size=20)))
    fig.update_yaxes(range=ylim, row=2, col=3, linecolor=line_color, zerolinecolor=line_color, gridcolor=line_color, ticksuffix='°', title=dict(text='Y', standoff=0, font=dict(size=20)))
    
    x_range = 150
    x_start = -2
    if subTR:
        x_start = 10000
        x_range *= item['real_y'].shape[1]
    fig.update_xaxes(range=[x_start, x_range+x_start], row=1, col=3, title=dict(text='Functional Volume (TR)', standoff=16, font=dict(size=20)))
    fig.update_xaxes(range=[x_start, x_range+x_start], row=2, col=3)

    return fig



import os
import pickle
import numpy as np
import ants
from os.path import join
from scipy.io import loadmat
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from deepmreye.util.data_io import download_mask

# --------------------------------------------------------------------------------
# --------------------------ANTS TRANSFORMS---------------------------------------
# --------------------------------------------------------------------------------
def register_to_eye_masks(dme_template, func, masks, verbose=1, transforms=None, metric='GC'):
    """Register functional to DeepMReye template (dme_template) using different sized masks

    Parameters
    ----------
    dme_template : ants Image
        Ants image with template file
    func : ants Image
        Functional image to register to dme_template
    masks : list
        List of Ants image objects containing variable sized masks
    verbose : int, optional
        Verbosity level of function, by default 1
    transforms : string, optional
        Which transforms should be used to transform image, by default None & set to Similarity
    metric : str, optional
        Which metric to quantify fit, by default 'GC'

    Returns
    -------
    func : ants Image
        Functional image registered to dme_template
    transformation_stats : array
        Statistics of transformation, used for dataset report
    """
    transformation_stats = []
    for idx, mask in enumerate(masks):
        if transforms is None:
            type_of_transform = 'Similarity'
        else:
            type_of_transform = transforms[idx]
        register_to_nau = ants.registration(fixed=dme_template, moving=func.get_average_of_timeseries(), aff_random_sampling_rate=1,
                                            type_of_transform=type_of_transform, mask=mask, aff_metric=metric, aff_sampling=512,
                                            aff_iterations=(200, 200, 200, 10), aff_smoothing_sigmas=(0, 0, 0, 0))
        if verbose > 0:
            if 'SyN' in type_of_transform:
                registered_fwd = loadmat(register_to_nau['fwdtransforms'][1])['AffineTransform_float_3_3']
            else:
                registered_fwd = loadmat(register_to_nau['fwdtransforms'][0])['AffineTransform_float_3_3']
            print('Mask {}/{}, Sum: {:.3f}, Mean {:.3f}, Std {:.3f}, Median {:.3f}'.format(idx, len(masks)-1, np.sum(registered_fwd),
                                                                           np.mean(registered_fwd), np.std(registered_fwd), np.median(registered_fwd)))
        transformation_stats.append(np.mean(registered_fwd))
        # Transform
        func = ants.apply_transforms(fixed=dme_template, moving=func,
                                     transformlist=register_to_nau['fwdtransforms'], imagetype=3)
            
    return func, np.array(transformation_stats)


def run_participant(fp_func, dme_template, eyemask_big, eyemask_small, x_edges, y_edges, z_edges, replace_with=0, transforms=None):
    """Run preprocessing for one participant with templates and masks preloaded to avoid computational overhead

    Parameters
    ----------
    fp_func : string
        Filepath to participant functional
    dme_template : ants Image
        Preloaded Image to dme_template
    eyemask_big : ants Image
        Big eyemask as ants Image
    eyemask_small : ants Image
        Small eyemask as ants Image
    x_edges : list
        Edges of mask in x-dimension
    y_edges : list
        Edges of mask in y-dimension
    z_edges : list
        Edges of mask in z-dimension
    replace_with : int, optional
        Values outside of mask are set to this, by default 0
    """
    # Load subject specific run. File should be Nifti and 4D but should also work with other formats which can be read with AntsPy
    func = ants.image_read(fp_func)
    # Register to deepmreye template (dme_template). If registration fails quality check, try below line with additional parameter "transforms=['Affine', 'Affine', 'SyNAggro']"
    transform_to_dme, transformation_statistics = register_to_eye_masks(dme_template, func, masks=[None, eyemask_big, eyemask_small], transforms=transforms)
    # Cut mask and save to subject folder with subject report / quality control plots
    (original_input, masked_eye, mask) = cut_mask(transform_to_dme, eyemask_small.numpy(), x_edges, y_edges, z_edges, replace_with=replace_with, save_overview=True, fp_func=fp_func)

    return (masked_eye, transformation_statistics)

# --------------------------------------------------------------------------------
# --------------------------MASKING-----------------------------------------------
# --------------------------------------------------------------------------------
def get_masks(data_path=''):
    """Loads masks for whole brain, big eye mask and small eye mask

    Parameters
    ----------
    data_path : str, optional
        Path to where masks are stored in .nii format, by default '../deepmreye/masks/'

    Returns
    -------
    eyemask_small : ants Image
        Eyemask containing voxels within the eye
    eyemask_big : ants Image
        Square eye mask centered on both eyes
    dme_template : ants Image
        Template brain using centered gaze positions
    mask : ants Image
        Mask which is used to cut 3D shape for model (in this case the same as eyemask_small)
    x_edges : list
        Edges of mask in x-dimension
    y_edges : list
        Edges of mask in y-dimension
    z_edges : list
        Edges of mask in z-dimension
    """     
    def load_from_path(fn_mask):
        if os.path.exists(fn_mask):
            return ants.image_read(fn_mask)
        else:
            print('Downloading mask: {}'.format(fn_mask))
            download_mask(fn_mask)
            return ants.image_read(fn_mask)

    if data_path == "":
        data_path = os.path.abspath(join(__file__, "..", "masks"))

    eyemask_small = load_from_path(join(data_path, 'eyemask_small.nii'))
    eyemask_big = load_from_path(join(data_path,  'eyemask_big.nii'))
    dme_template = load_from_path(join(data_path, 'dme_template.nii'))
    (mask, x_edges, y_edges, z_edges) = get_mask_edges(mask=eyemask_small)
    return eyemask_small, eyemask_big, dme_template, mask, x_edges, y_edges, z_edges


def get_mask_edges(mask, split=True):
    """Gets edges of mask

    Parameters
    ----------
    fp_mask : filepath, optional
        Filepath to mask
    split : bool, optional
        Splits masks into hemispheres, by default True

    Returns
    -------
    mask:
        Array of extracted mask edges
    x_edges, y_edges, z_edges: 
        Edges in (x,y,z)-dimension
    """
    # Get indices for left and right eye separately
    edge_indices = np.where(mask.numpy() == 1)
    if split:
        # Get left and right based on middle between left and right eye. For collin27 : 45
        middle_cut = (edge_indices[0][np.argmax(np.diff(edge_indices[0]))] +
                      edge_indices[0][np.argmax(np.diff(edge_indices[0])) + 1]) // 2
        # Get x and y values for both eyes and combine in one volume
        left_indices = edge_indices[0][edge_indices[0] < middle_cut]
        right_indices = edge_indices[0][edge_indices[0] > middle_cut]
        x_edges = (np.max(left_indices), np.min(left_indices), np.max(right_indices), np.min(right_indices))
    else:
        middle = np.min(edge_indices[0]) + (np.max(edge_indices[0]) - np.min(edge_indices[0])) // 2
        x_edges = (middle, np.min(edge_indices[0]), np.max(edge_indices[0]), middle)
    y_edges = (np.max(edge_indices[1]), np.min(edge_indices[1]))
    z_edges = (np.max(edge_indices[2]), np.min(edge_indices[2]))

    return (mask.numpy(), x_edges, y_edges, z_edges)


def cut_mask(to_mask, mask, x_edges, y_edges, z_edges, replace_with=0, save_overview=True, fp_func=None, verbose=0):
    """Cut mask into given shape given edges

    Parameters
    ----------
    to_mask : ants Image
        Image to mask
    mask : ants Image
        Mask as numpy array
    x_edges : list
        Edges of mask in x-dimension
    y_edges : list
        Edges of mask in y-dimension
    z_edges : list
        Edges of mask in z-dimension
    replace_with : int, optional
        Values outside of mask are set to this, by default 0
    save_overview : bool, optional
        Saves report / quality control figure when set to True, by default True
    fp_func : str, optional
        Filepath to new functional, by default None
    verbose : int, optional
        Verbosity level of this function, by default 0

    Returns
    -------
    original_input : ants Image
        Returns to_mask
    masked_eye : ants Image
        masked_eye as numpy array
    mask : ants Image
        Return mask
    """
    # Mask image to set out of mask values
    original_input = to_mask.copy()
    to_mask[mask < 1, ...] = replace_with
    # Slice for mask
    masked_eye_left = to_mask[x_edges[1]: x_edges[0], y_edges[1]: y_edges[0], z_edges[1]: z_edges[0], ...]
    masked_eye_right = to_mask[x_edges[3]: x_edges[2], y_edges[1]: y_edges[0], z_edges[1]: z_edges[0], ...]
    masked_eye = np.concatenate((masked_eye_right, masked_eye_left))
    if verbose > 0:
        print('Voxels > 0 / Mean of voxels: {} / {}'.format(np.sum(np.mean(masked_eye, axis=3) > 0), np.mean(masked_eye)))
    # Save back masked func to .nii and masked eye to .p
    participant_basename = os.path.basename(fp_func).split('.')[0]
    if save_overview:
        fn_full_mask = join(os.path.dirname(fp_func), f'report_{participant_basename}')
        plot_subject_report(fn_full_mask, original_input, masked_eye, mask)
    fn_masked_eye = join(os.path.dirname(fp_func), f'mask_{participant_basename}.p')
    pickle.dump(masked_eye, open(fn_masked_eye, 'wb'))

    return (original_input, masked_eye, mask)

# --------------------------------------------------------------------------------
# --------------------------VISUALIZATIONS----------------------------------------
# --------------------------------------------------------------------------------

def plot_subject_report(fn_subject, original_input, masked_eye, mask, color="rgb(0, 150, 175)", bg_color="rgb(14, 17, 23, 0)"):
    """Plots quality check figure for given subject

    Parameters
    ----------
    fn_subject : string
        Filepath to subject
    original_input : ants Image
        Filepath to functional image of subject
    masked_eye : array
        Numpy array of masked eye
    mask : ants Image
        ants mask
    color : str, optional
        Boxplot color, by default "rgb(0, 150, 175)"
    bg_color : str, optional
        Background color, by default "rgb(0,0,0)"
    """
    # Prepare data
    whole_brain_mask = original_input.get_average_of_timeseries()
    eye_mask = np.mean(masked_eye, axis=3)
    eye_mask_flat = eye_mask.flatten()
    # Also remove zero for histogram
    eye_mask_flat = eye_mask_flat[eye_mask_flat > 0]

    whole_brain_timecourse = np.mean(original_input.numpy(), axis=(0,1,2))
    eye_mask_timecourse = np.mean(masked_eye, axis=(0,1,2))

    # Normalize
    eye_mask_timecourse = (eye_mask_timecourse - np.min(eye_mask_timecourse)) / (np.max(eye_mask_timecourse) - np.min(eye_mask_timecourse))
    whole_brain_timecourse = (whole_brain_timecourse - np.min(whole_brain_timecourse)) / (np.max(whole_brain_timecourse) - np.min(whole_brain_timecourse))    
    
    # Plot
    fig = make_subplots(rows=2, cols=4, column_widths=[0.2, 0.2, 0.2, 0.4], row_heights=[0.6, 0.4], vertical_spacing=0.13)

    fig.add_trace(go.Heatmap(z=whole_brain_mask[25,:,:].transpose(), showscale=False, colorscale='Greys_r'),row=1, col=1)
    fig.add_trace(go.Heatmap(z=mask[25,:,:].transpose(), showscale=False, colorscale=[[0, 'rgba(0, 0, 0, 0)'], [1.0, 'rgba(255, 0, 0, 0.25)']]),row=1, col=1)
    fig.add_trace(go.Heatmap(z=whole_brain_mask[:,90,:].transpose(), showscale=False, colorscale='Greys_r'),row=1, col=2)
    fig.add_trace(go.Heatmap(z=mask[:,90,:].transpose(), showscale=False, colorscale=[[0, 'rgba(0, 0, 0, 0)'], [1.0, 'rgba(255, 0, 0, 0.25)']]),row=1, col=2)
    fig.add_trace(go.Heatmap(z=whole_brain_mask[:,:,15], showscale=False, colorscale='Greys_r'),row=1, col=3)
    fig.add_trace(go.Heatmap(z=mask[:,:,15], showscale=False, colorscale=[[0, 'rgba(0, 0, 0, 0)'], [1.0, 'rgba(255, 0, 0, 0.25)']]),row=1, col=3)

    fig.add_trace(go.Histogram(x=eye_mask_flat, nbinsx=75, marker = {'line' : {'width' : 0.75, 'color' : 'rgb(255, 255, 255)'}}, marker_color=color),row=1, col=4)

    fig.add_trace(go.Heatmap(z=np.mean(eye_mask, axis=0).transpose(), showscale=False, colorscale='Hot'),row=2, col=1)
    fig.add_trace(go.Heatmap(z=np.mean(eye_mask, axis=1).transpose(), showscale=False, colorscale='Hot'),row=2, col=2)
    fig.add_trace(go.Heatmap(z=np.mean(eye_mask, axis=2), showscale=False, colorscale='Hot'),row=2, col=3)
    fig.add_trace(go.Scatter(x=np.arange(0, len(eye_mask_timecourse)), y=eye_mask_timecourse, marker_color=color),row=2, col=4)
    fig.add_trace(go.Scatter(x=np.arange(0, len(whole_brain_timecourse)), y=whole_brain_timecourse, marker_color='rgb(255, 255, 255)'),row=2, col=4)

    annotations = [dict(x=0.07, y=1.03, xref='paper', yref='paper', text="x=-20", font=(dict(size=20)), showarrow=False),
                  dict(x=0.29, y=1.03, xref='paper', yref='paper', text="y=36", font=(dict(size=20)), showarrow=False),
                  dict(x=0.54, y=1.03, xref='paper', yref='paper', text="z=-30", font=(dict(size=20)), showarrow=False),
                  dict(x=0.17, y=1.1, xref='paper', yref='paper', text="<b>Transformed MNI space with eye mask (r)</b>", font=(dict(size=20)), showarrow=False),
                  dict(x=0.93, y=1.1, xref='paper', yref='paper', text="<b>Histogram of eye mask voxels</b>", font=(dict(size=20)), showarrow=False),
                  dict(x=0.24, y=0.42, xref='paper', yref='paper', text="<b>Eye mask voxels</b>", font=(dict(size=20)), showarrow=False),
                  dict(x=0.99, y=0.41, xref='paper', yref='paper', text="<b>Average across whole brain (w) & eye mask (b)</b>", font=(dict(size=20)), showarrow=False)]

    fig.update_layout(autosize=False,showlegend=False, width=1400, height=600, margin=dict(t=70, l=30, b=50, r=30),
                          plot_bgcolor=bg_color, paper_bgcolor=bg_color, font = {'color' : '#FFFFFF', 'size' : 13}, annotations=annotations)
    fig.update_xaxes(showgrid=False, showticklabels=True, col=4)
    fig.update_yaxes(showgrid=False, showticklabels=True, col=4)
    fig.update_yaxes(showgrid=False, showticklabels=True, showline=True, col=4, row=1)
    # # Remove labels from brain plots
    fig.update_yaxes(showgrid=False, showticklabels=True, col=1)
    fig.update_yaxes(showgrid=False, showticklabels=True, col=2)
    fig.update_yaxes(showgrid=False, showticklabels=True, col=3)
    fig.update_xaxes(showgrid=False, showticklabels=True, col=1)
    fig.update_xaxes(showgrid=False, showticklabels=True, col=2)
    fig.update_xaxes(showgrid=False, showticklabels=True, col=3)
    # Add mean and median to hist
    fig.add_vline(x=np.mean(eye_mask_flat), annotation=dict(text='Mean', y=0.9), line=dict(color='rgb(255, 255, 255)'), row=1, col=4)
    fig.add_vline(x=np.median(eye_mask_flat), annotation=dict(text='Median'), line=dict(color='rgb(255, 255, 255)'), row=1, col=4)
    
    fig.write_html(fn_subject + ".html")
    
# --------------------------------------------------------------------------------
# -----------------------IMG MANIPULATIONS----------------------------------------
# --------------------------------------------------------------------------------

def normalize_img(img_in, mad_time=False, standardize_tr=True, std_cut_after=5):
    """As part of preprocessing the 4D input is normalized across different dimensions

    Parameters
    ----------
    img_in : ants Image
        Image to normalize
    mad_time : bool, optional
        Determines if median absolute deviation should be used across time dimension, by default False
    standardize_tr : bool, optional
        Determines if each image should be normalized across spatial dimensions, by default True
    std_cut_after : int, optional
        Gets rid of outliers after normalization, by default 5

    Returns
    -------
    img_in : ants Image
        Normalized output image
    """    
    # Transpose so time comes first
    img_in = np.transpose(img_in, axes=(3, 0, 1, 2))
    zero_indices = img_in == 0
    img_in[zero_indices] = np.NaN

    # Median absolute deviation (MAD)
    if mad_time:
        est_mean = np.nanmedian(img_in, axis=0)
        est_std = np.nanmedian(abs(img_in - est_mean), axis=0)
        img_in = (img_in - est_mean) / est_std
    else:
        img_in = (img_in - np.nanmean(img_in, axis=0)) / np.nanstd(img_in, axis=0)

    # Normalize each functional on its own:
    if standardize_tr:
        img_in = np.array([(x - np.nanmean(x)) / np.nanstd(x) for x in img_in])
    if std_cut_after is not None:
        std_before = np.nanstd(img_in)
        img_in[img_in > std_cut_after*std_before] = std_cut_after*std_before
        img_in[img_in < -std_cut_after*std_before] = -std_cut_after*std_before

    # If division by zero replace with 0
    img_in[~np.isfinite(img_in)] = 0

    # Transpose back to original
    img_in[zero_indices] = 0
    img_in = np.transpose(img_in, axes=(1, 2, 3, 0))

    return img_in

# --------------------------------------------------------------------------------
# --------------------------LABEL I/O---------------------------------------------
# --------------------------------------------------------------------------------

def load_label(label_path, label_type='calibration_run'):
    """Load label for experiment, which should return X,Y coordinates for each timepoint.
    This function can be exchanged for experiment specific loading of labels, or by using different label types.

    Parameters
    ----------
    label_path : str
        Path to file with labels
    label_type : str, optional
        Which type of labels are used in the experiment, by default 'calibration_run'

    Returns
    -------
    this_label : numpy array
        X,Y coordinates for each functional describing gaze position during this timepoint. 
    """    
    if label_type=='calibration_run':
        # Load labels from file
        fn_labels = join(label_path, 'stim_vals.csv')
        labels = np.genfromtxt(fn_labels, delimiter=',')
        labels = labels[1:]
        labels = np.repeat(labels, 5, axis=0)
        this_label = labels[:, np.newaxis, :]
        this_label = np.repeat(this_label, 10, axis=1)

        # Normalize label
        this_label = (this_label - -0.95) / ( 0.95 - -0.95)
        this_label -= 0.5

        # Y-axis is flipped for this dataset
        this_label[...,1] *= -1

        # Convert to visual angles
        this_label[...,0] *= 19
        this_label[...,1] *= 14.7
    
    return this_label

def save_data(participant, participant_data, participant_labels, participant_ids, processed_data, center_labels=False):
    """Save participant data to npz file for fast (lazy) loading during model training

    Parameters
    ----------
    participant : str
        Participant label
    participant_data : list
        4D (X,Y,Z,t) data for participant across runs
    participant_labels : list
        3D (t,X,Y) data, with corresponding labels to participant data
    participant_ids : str
        Participant identifier with run id
    processed_data : str
        Filepath to where processed data should be stored
    center_labels : bool, optional
        Centers labels to (0,0) which can improve performance of model, by default False
    """
    # Save npz file for each participant. The resulting file contains both the eye mask and the labels which are used for model training
    participant_data = np.transpose(np.concatenate(participant_data, axis=3), axes=(3,0,1,2))
    participant_labels = np.concatenate(participant_labels, axis=0).astype('float32')
    participant_ids = np.concatenate(participant_ids, axis=1).transpose()

    # Adjust labels to be centered at (0,0)
    if center_labels:
        participant_labels = participant_labels - np.nanmedian(participant_labels, axis=(0,1))
        
    # Save data and labels in npz file (lazy loading)
    data_dict = dict()
    for idx, (data, label, identifier) in enumerate(zip(participant_data, participant_labels, participant_ids)):
        data_dict['data_{}'.format(idx)] = data
        data_dict['label_{}'.format(idx)] = label
        data_dict['identifier_{}'.format(idx)] = identifier

    # Save each subject in separate .npz files (fast to load)
    subject_file_path = join(processed_data, participant)
    print('Saving eye data {} and targets {} (NaN {}) to file {}'.format(participant_data.shape, participant_labels.shape, np.sum(np.isnan(participant_labels.flatten())), subject_file_path))
    np.savez(subject_file_path, **data_dict)



import random
import os
import numpy as np
from pprint import pprint
from . import util


def create_generators(full_training_list, full_testing_list, batch_size=8, withinsubject_split=None, augment_list=[0, 0, 0], mixed_batches=True, inner_timesteps=None):
    # Overall training and testing generator
    training_generator = data_generator(full_training_list, batch_size, training=True, inner_timesteps=inner_timesteps,
                                        augment_list=augment_list, mixed_batches=mixed_batches, withinsubject_split=withinsubject_split)
    testing_generator = data_generator(full_testing_list, batch_size, training=False, inner_timesteps=inner_timesteps,
                                       mixed_batches=mixed_batches, withinsubject_split=withinsubject_split)

    # For each testing subject create a single generator
    single_testing_generators, single_testing_names = get_single_data_generators(full_testing_list, batch_size, string_cut=7, training=False,
                                                                                 mixed_batches=mixed_batches, withinsubject_split=withinsubject_split, inner_timesteps=inner_timesteps)
    single_training_generators, single_training_names = get_single_data_generators(full_training_list, batch_size, string_cut=7, training=True,
                                                                                   augment_list=augment_list, mixed_batches=mixed_batches, withinsubject_split=withinsubject_split, inner_timesteps=inner_timesteps)

    training_subjects_string = [os.path.splitext(os.path.basename(p))[0] for p in full_training_list]
    test_subjects_string = [os.path.splitext(os.path.basename(p))[0] for p in full_testing_list]

    print(util.color.BOLD + 'Training set ({}) contains {} subjects: '.format(
        os.path.dirname(full_training_list[0]), len(full_training_list)) + util.color.END)
    pprint(training_subjects_string, compact=True)
    print(util.color.BOLD + 'Test set ({}) contains {} subjects: '.format(
        os.path.dirname(full_testing_list[0]), len(full_testing_list)) + util.color.END)
    pprint(test_subjects_string, compact=True)
    # Should be a dict.
    return (training_generator, testing_generator, single_testing_generators, single_testing_names, single_training_generators, single_training_names)


def create_holdout_generators(datasets, train_split=0.6, **args):
    full_training_list, full_testing_list = list(), list()
    for fn_data in datasets:
        this_file_list = [fn_data + p for p in os.listdir(fn_data)]
        np.random.shuffle(this_file_list)
        train_test_split = int(train_split * len(this_file_list))
        this_training_list = this_file_list[0:train_test_split]
        this_testing_list = this_file_list[train_test_split::]
        full_training_list.extend(this_training_list)
        full_testing_list.extend(this_testing_list)
    (training_generator, testing_generator, single_testing_generators, single_testing_names, single_training_generators,
     single_training_names) = create_generators(full_training_list, full_testing_list, **args)

    return (training_generator, testing_generator, single_testing_generators, single_testing_names,
            single_training_generators, single_training_names, full_testing_list, full_training_list)


def create_cv_generators(dataset, num_cvs=5, **args):
    # Create lists for each cv
    this_file_list = [dataset + p for p in os.listdir(dataset)]
    np.random.shuffle(this_file_list)
    cv_split = np.array_split(this_file_list, num_cvs)
    cv_return = []
    for idx, cvs in enumerate(cv_split):
        full_testing_list = cvs.tolist()
        full_training_list = np.concatenate([x for i, x in enumerate(cv_split) if i != idx]).tolist()

        (training_generator, testing_generator,
         single_testing_generators, single_testing_names,
         single_all_generators, single_all_names) = create_generators(full_training_list, full_testing_list, **args)
        cv_return.append((training_generator, testing_generator, single_testing_generators, single_testing_names,
                          single_all_generators, single_all_names, full_testing_list, full_training_list))

    return cv_return


def create_leaveoneout_generators(datasets, training_subset=None, **args):
    loo_return = []
    for idx, dataset in enumerate(datasets):
        full_testing_list = [dataset + p for p in os.listdir(dataset)]
        training_datasets = [x for i, x in enumerate(datasets) if i != idx]
        full_training_list = [tds + p for tds in training_datasets for p in os.listdir(tds)]

        if training_subset is not None:
            size_before = len(full_training_list)
            full_training_list = [tds for tds in full_training_list if os.path.basename(
                os.path.dirname(tds)) + '/' + os.path.basename(tds) in training_subset]
            print('Using subset ({} / {}) for training {}'.format(len(full_training_list), size_before, dataset))

        (training_generator, testing_generator,
         single_testing_generators, single_testing_names,
         single_all_generators, single_all_names) = create_generators(full_training_list, full_testing_list, **args)
        loo_return.append((training_generator, testing_generator, single_testing_generators, single_testing_names,
                           single_all_generators, single_all_names, full_testing_list, full_training_list))

    return loo_return


def get_single_data_generators(fn_list, batch_size, string_cut=4, **args):
    generators, names = list(), list()
    for subject in fn_list:
        this_generator = data_generator([subject], batch_size, **args)
        generators.append(this_generator)
        this_name = os.path.basename(subject)[:-4]
        if len(this_name) > string_cut:
            this_name = this_name[-string_cut:]
        names.append(this_name)
    return generators, names


def data_generator(file_list, batch_size, training=False, mixed_batches=True,
                   withinsubject_split=None, augment_list=[0, 0, 0], inner_timesteps=None):
    """
    Take a random subject, load it and return a batched subset    
    """
    all_nonan_indices = get_nonan_indices(file_list)
    start_tr, end_tr = get_start_end_tr(withinsubject_split, training)
    if withinsubject_split is None:
        file_list = [file_list[x] for x in range(0, len(all_nonan_indices)) if all_nonan_indices[x].size > 0]
        all_nonan_indices = [all_nonan_indices[x] for x in range(0, len(all_nonan_indices)) if all_nonan_indices[x].size > 0]
    else:
        new_filelist, new_nonan = [], []
        for x in range(0, len(all_nonan_indices)):
            try:
                get_subject_data(file_list[x], batch_size=1, start_tr=start_tr,
                                 end_tr=end_tr, nonan_indices=all_nonan_indices[x])
                new_filelist.append(file_list[x])
                new_nonan.append(all_nonan_indices[x])
            except ValueError:
                pass
        file_list = new_filelist
        all_nonan_indices = new_nonan
    while True:
        if mixed_batches:
            fn_subjects, nonan_indices = zip(*random.choices(list(zip(file_list, all_nonan_indices)), k=batch_size))
            X, y = [], []
            for subj, nans in zip(fn_subjects, nonan_indices):
                tmp_X, tmp_y = get_subject_data(subj, batch_size=1, start_tr=start_tr,
                                                end_tr=end_tr, nonan_indices=nans)
                X.append(tmp_X[0])
                y.append(tmp_y[0])
        else:
            fn_subject = np.random.choice(file_list)
            X, y = get_subject_data(fn_subject, batch_size=batch_size, start_tr=start_tr, end_tr=end_tr)
        # Downsample to number of inner timesteps (subTR)
        if inner_timesteps is not None:
            y = [y_tmp[np.linspace(0, y[0].shape[0] - 1, inner_timesteps, dtype=int), :] for y_tmp in y]
        X, y = np.array(X)[..., np.newaxis], np.array(y)
        # Augmentation
        if training and any(augment_list):
            X = util.augment_input(X, rotation=augment_list[0], shift=augment_list[1], zoom=augment_list[2])

        yield [[X, y], []]


def get_subject_data(fn_subject, batch_size=None, sample_index=None, start_tr=None, end_tr=None, nonan_indices=None):
    data = np.load(fn_subject, mmap_mode='r')
    if batch_size is not None:
        try:
            _ = data['identifier_0']
            divisor = 3
        except:
            divisor = 2
        num_trs = len(data) // divisor
        start_tr, end_tr = get_tr_indices(num_trs, start_tr, end_tr)
        nonan_indices = nonan_indices[(nonan_indices >= start_tr) & (nonan_indices <= end_tr)]
        batch_indices = np.random.choice(nonan_indices, batch_size)
        X = [data['data_' + str(b)] for b in batch_indices]
        y = [data['label_' + str(b)] for b in batch_indices]
    elif sample_index is not None:
        X = data['data_' + str(sample_index)]
        y = data['label_' + str(sample_index)]
    else:
        return data
    return X, y


def get_tr_indices(num_trs, start_tr, end_tr):
    if start_tr is None:
        start_tr = 0
    else:
        start_tr = int(start_tr * num_trs)
    if end_tr is None:
        end_tr = num_trs
    else:
        end_tr = int(end_tr * num_trs)
    return (start_tr, end_tr)


def get_start_end_tr(withinsubject_split, training):
    if withinsubject_split:
        if training:
            start_tr = withinsubject_split[0]
            end_tr = withinsubject_split[1]
        else:
            if withinsubject_split[0] == 0:
                start_tr = withinsubject_split[1]
                end_tr = None
            else:
                start_tr = None
                end_tr = withinsubject_split[0]
    else:
        start_tr = None
        end_tr = None
    return (start_tr, end_tr)


def get_nonan_indices(file_list):
    nonan_indices = []
    for fn_subject in file_list:
        data = np.load(fn_subject, mmap_mode='r')
        try:
            _ = data['identifier_0']
            divisor = 3
        except:
            divisor = 2
        num_trs = len(data) // divisor
        y = np.array([data['label_' + str(sample_index)] for sample_index in range(0, num_trs)])
        nonan_index = ~np.any(np.isnan(y), axis=(1, 2))
        nonan_indices.append(np.where(nonan_index)[0])
    return nonan_indices


def get_all_subject_data(fn_subject):
    data = np.load(fn_subject, mmap_mode='r')
    try:
        _ = data['identifier_0']
        divisor = 3
    except:
        divisor = 2
    num_trs = len(get_subject_data(fn_subject)) // divisor
    subject_data_X, subject_data_y = [], []
    for sample_index in range(0, num_trs):
        subject_data_X.append(data['data_' + str(sample_index)])
        subject_data_y.append(data['label_' + str(sample_index)])
    subject_data_X, subject_data_y = np.array(subject_data_X), np.array(subject_data_y)

    return (subject_data_X[..., np.newaxis], subject_data_y)



#!/usr/bin/env python
"""
DeepMReye Toolbox
© Markus Frey, Matthias Nau
https://github.com/DeepMReye/DeepMReye
Licensed under LGPL-3.0 License
"""
import sys

from setuptools import setup


SETUP_REQUIRES = ['setuptools >= 30.3.0']
SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []

if __name__ == "__main__":
    setup(setup_requires=SETUP_REQUIRES)



from . import util


def get_opts():
    opts = dict()

    opts['kernel'] = 3 # Kernel size for all convolutional layers
    opts['lr'] = 0.00002 # Learning rate
    opts['filters'] = 32 # Number of filters in convolutional layers
    opts['multiplier'] = 2 # How much does the number of filters increase in each layer
    opts['depth'] = 4 # Maximum number of layers
    opts['dropout_rate'] = 0.1 # Dropout ratio for fully connected layers
    opts['num_dense'] = 2 # Number of fully connected layers
    opts['num_fc'] = 1024 # Number of units in fully connected layer
    opts['gaussian_noise'] = 0 # How much gaussian noise is added (unit = standard deviation)
    opts['activation'] = util.mish # Activation function for all layers
    opts['groups'] = 8 # Number of groups to normalize across (see GroupNorm)
    opts['inner_timesteps'] = 10 # Default number of subTR samples which are being reconstructed

    # Loss weights
    opts['loss_euclidean'] = 1 # Loss weight for euclidean distance
    opts['loss_confidence'] = 0.1 # Loss weight for uncertainty measure

    # Training
    opts['epochs'] = 25 # Number of epochs
    opts['steps_per_epoch'] = 1500 # Number of steps per training epoch
    opts['validation_steps'] = 1500 # Number of steps per validation epoch
    opts['train_test_split'] = 0.6 # Default proportion of train (60%)-test(40%) split
    opts['batch_size'] = 8 # Batch size used for training the model
    opts['mixed_batches'] = True # If true, each batch contains samples across participants
    opts['mc_dropout'] = False # If true, monte carlo dropout is used
    opts['rotation_x'] = 5 # Augmentation parameter, rotation in x-axis
    opts['rotation_y'] = 5 # Augmentation parameter, rotation in y-axis
    opts['rotation_z'] = 5 # Augmentation parameter, rotation in z-axis
    opts['shift'] = 4 # Augmentation parameter, shift in all axes
    opts['zoom'] = 0.15 # Augmentation parameter, zoom in all axes

    return opts



import os
import urllib
import numpy as np
from scipy.io import loadmat


# --------------------------------------------------------------------------------
# --------------------------IO-NAU------------------------------------------------
# --------------------------------------------------------------------------------
def get_subject_labels(subject_string, mat_data):
    for subject in mat_data:
        if str(subject['subID']) == subject_string:
            all_runs = subject['TR_xy'].tolist()
            return all_runs
    return -1


def get_all_subject_labels(subject_string, mat_data, num_downsampled=10, use_real=False):
    """
    For models with multiple outputs we want to estimate the sub-TR XY 
    Inputs: 
        - subject_string : Subject identified
        - mat_data : Data to subject logs in mat format
        - num_downsampled : How many sub-TR XY are left in the output
    """
    real_et = None
    for subject in mat_data:
        if str(subject['subID']) == subject_string:
            try:
                if use_real:
                    all_runs = subject['TR_xy_samples_ET'].tolist()
                    real_et = True
                else:
                    raise ValueError('Hack')
            except ValueError:
                print('Subject {} has no real eye tracking, use XY of moving dot.'.format(subject_string))
                all_runs = subject['TR_xy_samples'].tolist()
                real_et = False
            subject_xy = list()
            for run in all_runs:
                run_xy = list()
                for ft in run:
                    if len(ft.shape) > 1:
                        ft_xy = ft[:, np.linspace(0, ft.shape[1] - 1, num_downsampled, dtype=int)]
                        run_xy.append(ft_xy.transpose())
                    else:
                        run_xy.append(np.full([num_downsampled, 2], np.nan))
                subject_xy.append(np.array(run_xy))
            subject_xy = np.array(subject_xy)
            return (subject_xy, real_et)
    return (-1, -1)

# --------------------------------------------------------------------------------
# --------------------------IO-IGN------------------------------------------------
# --------------------------------------------------------------------------------


def get_all_subject_labels_ign(subject_string, num_downsampled=10):
    """
    For models with multiple outputs we want to estimate the sub-TR XY 
    Inputs: 
        - subject_string : Subject identified
        - mat_data : Data to subject logs in mat format
        - num_downsampled : How many sub-TR XY are left in the output
    """
    subj_data = np.load(subject_string)
    np.testing.assert_(np.unique(np.diff(np.where(subj_data[:, 0] == 1)[0]))[0] == 510)

    all_labels, all_runs = [], []
    for idx, l in enumerate(subj_data):
        if l[1] == 0:
            if all_labels:
                all_runs.append(np.array(all_labels))
            all_labels = []
        bln_tr = int(l[0])
        if bln_tr:
            if (idx + 510) >= subj_data.shape[0]:
                this_label = subj_data[np.linspace(idx, subj_data.shape[0] - 1, num_downsampled, dtype=int), 2:4]
            else:
                this_label = subj_data[np.linspace(idx, idx + 510, num_downsampled, dtype=int), 2:4]
            all_labels.append(this_label)
    all_runs.append(np.array(all_labels))
    all_runs = np.array(all_runs)

    return all_runs

# --------------------------------------------------------------------------------
# --------------------------IO-BMD------------------------------------------------
# --------------------------------------------------------------------------------


def get_all_subject_labels_bmd(subject_string, run_idx, num_downsampled=10, real_et=False):
    """
    For models with multiple outputs we want to estimate the sub-TR XY 
    Inputs: 
        - subject_string : Subject identified
        - run_idx : Index for run
        - num_downsampled : How many sub-TR XY are left in the output
    """
    mat_data = loadmat(subject_string, mat_dtype=True)
    if real_et:
        mat_data = mat_data['XY']['samples_ET'][0, 0]
    else:
        mat_data = mat_data['XY']['samples'][0, 0]
    if run_idx < mat_data.shape[1]:
        if mat_data[0, run_idx].shape[0] == 0:
            return np.array([])
        this_run = mat_data[0, run_idx][0, :]
        all_subtr = np.array([x[np.linspace(0, len(x) - 1, num_downsampled, dtype=int), :]
                              if x.size > 0 else np.zeros((num_downsampled, 2)) * np.nan for x in this_run])
        return all_subtr
    return np.array([])

# --------------------------------------------------------------------------------
# --------------------------IO-MMD------------------------------------------------
# --------------------------------------------------------------------------------


def get_all_subject_labels_mmd(subject_string, run_idx, num_downsampled=10):
    """
    For models with multiple outputs we want to estimate the sub-TR XY 
    Inputs: 
        - subject_string : Subject identified
        - run_idx : Index for run
        - num_downsampled : How many sub-TR XY are left in the output
    """
    mat_data = loadmat(subject_string, mat_dtype=True)
    mat_data = mat_data['XY']['samples_ET'][0, 0]
    if run_idx < mat_data.shape[1]:
        this_run = mat_data[0, run_idx][0, :]
        all_subtr = np.array([x[np.linspace(0, len(x) - 1, num_downsampled, dtype=int), :]
                              if x.size > 0 else np.zeros((num_downsampled, 2)) * np.nan for x in this_run])
        return all_subtr
    return np.array([])


# --------------------------------------------------------------------------------
# --------------------------IO-MASKS----------------------------------------------
# --------------------------------------------------------------------------------


def download_mask(data_path, remote_path='https://github.com/DeepMReye/DeepMReye/blob/main/deepmreye/masks/'):
    mask_name = os.path.basename(data_path)
    mask_remote = remote_path + '{}?raw=true'.format(mask_name)
    try:
        (f, m) = urllib.request.urlretrieve(mask_remote, data_path)
    except urllib.error.URLError as e:
        raise RuntimeError("Failed to download '{}'. '{}'".format(mask_remote, e.reason))